@@ -6,16 +6,30 @@ Created by Kang Zhang on 2009-07-09
 
 import os
 import sys
-import getpass
 import ConfigParser
 
-from abc import ABCMeta, abstractmethod
+from keyring.util.escape import escape as escape_for_ini
+
+try:
+    from abc import ABCMeta, abstractmethod
+except ImportError:
+    # to keep compatible with older Python versions.
+    class ABCMeta(type):
+        pass
+
+    def abstractmethod(funcobj):
+        return funcobj
+
 
 _KEYRING_SETTING = 'keyring-setting'
 _CRYPTED_PASSWORD = 'crypted-password'
 _BLOCK_SIZE = 32
 _PADDING = '0'
 
+class PasswordSetError(Exception):
+    """Raised when the password can't be set.
+    """
+
 class KeyringBackend(object):
     """The abstract base class of the keyring, every backend must implement
     this interface.
@@ -35,13 +49,13 @@ class KeyringBackend(object):
     def get_password(self, service, username):
         """Get password of the username for the service
         """
-        pass
+        return None
 
     @abstractmethod
     def set_password(self, service, username, password):
         """Set password for the username of the service
         """
-        return -1
+        raise PasswordSetError()
 
 class _ExtensionKeyring(KeyringBackend):
     """_ExtensionKeyring is a adaptor class for the platform related keyring
@@ -79,17 +93,19 @@ class _ExtensionKeyring(KeyringBackend):
         try:
             password = self.keyring_impl.password_get(service, username)
         except OSError:
-            password = None 
+            password = None
         return password
 
     def set_password(self, service, username, password):
         """Overide the set_password() in KeyringBackend.
         """
-        return self.keyring_impl.password_set(service, username, password)
+        try:
+            self.keyring_impl.password_set(service, username, password)
+        except OSError:
+            raise PasswordSetError()
 
 class OSXKeychain(_ExtensionKeyring):
-    """The keyring backend based on Keychain Service of the Mac OSX
-    """
+    """Mac OSX Keychain"""
     def _init_backend(self):
         """Return the handler: osx_keychain
         """
@@ -102,8 +118,7 @@ class OSXKeychain(_ExtensionKeyring):
         return sys.platform == 'darwin'
 
 class GnomeKeyring(_ExtensionKeyring):
-    """The keyring backend using Gnome Keyring.
-    """
+    """Gnome Keyring"""
     def _init_backend(self):
         """Return the gnome_keyring handler.
         """
@@ -118,8 +133,7 @@ class GnomeKeyring(_ExtensionKeyring):
 
 
 class KDEKWallet(_ExtensionKeyring):
-    """The keyring backend based on KDE KWallet
-    """
+    """KDE KWallet"""
     def _init_backend(self):
         """Return the kde_kwallet handler.
         """
@@ -164,6 +178,9 @@ class BasicFileKeyring(KeyringBackend):
     def get_password(self, service, username):
         """Read the password from the file.
         """
+        service = escape_for_ini(service)
+        username = escape_for_ini(username)
+
         # load the passwords from the file
         config = ConfigParser.RawConfigParser()
         if os.path.exists(self.file_path):
@@ -183,6 +200,9 @@ class BasicFileKeyring(KeyringBackend):
     def set_password(self, service, username, password):
         """Write the password in the file.
         """
+        service = escape_for_ini(service)
+        username = escape_for_ini(username)
+
         # encrypt the password
         password_encrypted = self.encrypt(password)
         # load the password from the disk
@@ -199,11 +219,8 @@ class BasicFileKeyring(KeyringBackend):
         config_file = open(self.file_path,'w')
         config.write(config_file)
 
-        return 0
-
 class UncryptedFileKeyring(BasicFileKeyring):
-    """A simple filekeyring which dose not encrypt the password.
-    """
+    """Uncrypted File Keyring"""
     def filename(self):
         """Return the filename of the password file. It should be
         "keyring_pass.cfg" .
@@ -226,8 +243,7 @@ class UncryptedFileKeyring(BasicFileKeyring):
         return 0
 
 class CryptedFileKeyring(BasicFileKeyring):
-    """CryptedFileKeyring is a keyring using lib pycryto to encrypt the password
-    """
+    """PyCrypto File Keyring"""
     def __init__(self):
         super(CryptedFileKeyring, self).__init__()
         self.crypted_password = None
@@ -255,6 +271,7 @@ class CryptedFileKeyring(BasicFileKeyring):
         password = None
         while 1:
             if not password:
+                import getpass
                 password = getpass.getpass()
                 password2 = getpass.getpass('Password (again): ')
                 if password != password2:
@@ -316,6 +333,7 @@ class CryptedFileKeyring(BasicFileKeyring):
             self._init_file()
 
         print "Please input your password for the keyring"
+        import getpass
         password = getpass.getpass()
 
         if not self._auth(password):
@@ -343,9 +361,7 @@ class CryptedFileKeyring(BasicFileKeyring):
 
 
 class Win32CryptoKeyring(BasicFileKeyring):
-    """Win32CryptoKeyring is a keyring which use Windows CryptAPI to encrypt
-    the user's passwords and store them in a file.
-    """
+    """Win32 Cryptography Keyring"""
     def __init__(self):
         super(Win32CryptoKeyring, self).__init__()
 
@@ -361,14 +377,15 @@ class Win32CryptoKeyring(BasicFileKeyring):
         return "wincrypto_pass.cfg"
 
     def supported(self):
-        """Recommend for all Windows is higher than Windows 2000.
+        """Recommend when other Windows backends are unavailable
         """
-        if self.crypt_handler is not None and sys.platform == 'win32':
-            major, minor, build, platform, text = sys.getwindowsversion()
-            if platform == 2:
-                # recommend for windows 2k+
-                return 1
-        return -1
+        recommended = select_windows_backend()
+        if recommended == None:
+            return -1
+        elif recommended == 'file':
+            return 1
+        else:
+            return 0
 
     def encrypt(self, password):
         """Encrypt the password using the CryptAPI.
@@ -381,6 +398,132 @@ class Win32CryptoKeyring(BasicFileKeyring):
         return self.crypt_handler.decrypt(password_encrypted)
 
 
+class WinVaultKeyring(KeyringBackend):
+    def __init__(self):
+        super(WinVaultKeyring, self).__init__()
+        try:
+            import pywintypes, win32cred
+            self.win32cred = win32cred
+            self.pywintypes = pywintypes
+        except ImportError:
+            self.win32cred = None
+    
+    def supported(self):
+        '''Default Windows backend, when it is available
+        '''
+        recommended = select_windows_backend()
+        if recommended == None:
+            return -1
+        elif recommended == 'cred':
+            return 1
+        else:
+            return 0
+    
+    def get_password(self, service, username):
+        try:
+            blob = self.win32cred.CredRead(Type=self.win32cred.CRED_TYPE_GENERIC, 
+                                           TargetName=service)['CredentialBlob']
+        except self.pywintypes.error, e:
+            if e[:2] == (1168, 'CredRead'):
+                return None
+            raise
+        return blob.decode("utf16")
+    
+    def set_password(self, service, username, password):
+        credential = dict(Type=self.win32cred.CRED_TYPE_GENERIC, 
+                          TargetName=service,
+                          UserName=username,
+                          CredentialBlob=unicode(password),
+                          Comment="Stored using python-keyring",
+                          Persist=self.win32cred.CRED_PERSIST_ENTERPRISE)
+        self.win32cred.CredWrite(credential, 0)
+
+
+class Win32CryptoRegistry(KeyringBackend):
+    """Win32CryptoRegistry is a keyring which use Windows CryptAPI to encrypt
+    the user's passwords and store them under registry keys
+    """
+    def __init__(self):
+        super(Win32CryptoRegistry, self).__init__()
+
+        try:
+            import win32_crypto
+            import _winreg
+            self.crypt_handler = win32_crypto
+        except ImportError:
+            self.crypt_handler = None
+
+    def supported(self):
+        """Return if this keyring supports current enviroment.
+        -1: not applicable
+         0: suitable
+         1: recommended
+        """
+        recommended = select_windows_backend()
+        if recommended == None:
+            return -1
+        elif recommended == 'reg':
+            return 1
+        else:
+            return 0
+
+    def get_password(self, service, username):
+        """Get password of the username for the service
+        """
+        from _winreg import HKEY_CURRENT_USER, OpenKey, QueryValueEx
+        try:
+            # fetch the password
+            key = r'Software\%s\Keyring' % service
+            hkey = OpenKey(HKEY_CURRENT_USER, key)
+            password_base64 = QueryValueEx(hkey, username)[0]
+            # decode with base64
+            password_encrypted = password_base64.decode("base64")
+            # decrypted the password
+            password = self.crypt_handler.decrypt(password_encrypted)
+        except EnvironmentError:
+            password = None
+        return password
+
+
+    def set_password(self, service, username, password):
+        """Write the password to the registry
+        """
+        # encrypt the password
+        password_encrypted = self.crypt_handler.encrypt(password)
+        # encode with base64
+        password_base64 = password_encrypted.encode("base64")
+
+        # store the password
+        from _winreg import HKEY_CURRENT_USER, CreateKey, SetValueEx, REG_SZ
+        hkey = CreateKey(HKEY_CURRENT_USER, r'Software\%s\Keyring' % service)
+        SetValueEx(hkey, username, 0, REG_SZ, password_base64)
+
+def select_windows_backend():
+    if os.name != 'nt':
+        return None
+    major, minor, build, platform, text = sys.getwindowsversion()
+    try:
+        import pywintypes, win32cred
+        if (major, minor) >= (5, 1):
+            # recommend for windows xp+
+            return 'cred'
+    except ImportError:
+        pass
+    try:
+        import win32_crypto, _winreg
+        if (major, minor) >= (5, 0):
+            # recommend for windows 2k+
+            return 'reg'
+    except ImportError:
+        pass
+    try:
+        import win32_crypto
+        return 'file'
+    except ImportError:
+        pass
+    return None
+
+
 _all_keyring = None
 
 def get_all_keyring():
@@ -390,6 +533,7 @@ def get_all_keyring():
     if _all_keyring is None:
         _all_keyring = [ OSXKeychain(), GnomeKeyring(), KDEKWallet(),
                          CryptedFileKeyring(), UncryptedFileKeyring(),
-                         Win32CryptoKeyring()]
+                         Win32CryptoKeyring(), Win32CryptoRegistry(),
+                         WinVaultKeyring()]
     return _all_keyring
 
